program fdisk;

(* Create MBR and Partitions for CPM3, Armin Diehl 12/2006
   Based on gide code from Tillmann Reh *)

const  signon = ^m^j'IDE Harddisk Partition Utility V0.2  AD 25 Dec 2006'^m^j;

(* default geometry of connected harddisk *)
(* here: default mode IBM DTCA-24090, 4GB *)
(* will only be used as default, real values will be read from MBR *)

const  cylinders   : integer = 7944;
       heads       : integer = 16;
       sectors     : integer = 63;

(* I/O addresses and commands of the IDE interface/drive *)
(* The I/O addresses are user selectable in steps of $10 *)

       GIDEbase    : integer = $80;  (* GIDE base address *)

       cmd_readsector  = $20;
       cmd_writesector = $30;
       cmd_seek        = $70;
       cmd_diagnostics = $90;
       cmd_initialize  = $91;
       cmd_identify    = $EC;
       cmd_readbuffer  = $E4;
       cmd_writebuffer = $E8;

(* variables *)

type   workstr       = string[30];
       buftype       = array[0..511] of byte;
       str           = string[80];
       CPMPartDataT  = Record
                         cpmID   : array[0..2] of char;  (* CPM *)
                         reserved: array[0..5] of byte;
                         Tracks : integer;
                         Heads  : byte;
                         Sectors: byte;
                         BootPartNum : byte;
                         Partitions: array[0..15] of record
                                                       Start,Len: integer;
                                                     end;
                       end;
       DosPartEntryT = record
                         Boot   : byte;
                         SHead   : byte;
                         SSect   : byte; (* lower 6 bit=Sector, upper 2 bit=upper trk *)
                         STrk    : byte; (* lower 8 bit *)
                         SysId   : byte;
                         EHead   : byte;
                         ESect   : byte;
                         ETrk    : byte;
                         RelStartSecL,
                         RelStartSecH,
                         RelSecsL,
                         RelSecsH : integer;
                       end;

       MBRT = Record  (* track 0, head 0, sector 1 *)
                BootCode : array [0..367] of byte;
                PartData : CPMPartDataT;
                DosPartData : array [0..3] of DosPartEntryT;
                Id55 : byte;
                IdAA : byte;
              end;
      BootT = record  (* first sector of a partition *)
                 id : array[0..5] of char;  (* CPM3AD *)
                 SPT : integer;  (* 128 byte recs per track(all heads) *)
                 BSH : byte;
                 BLM : byte;
                 EXM : byte;
                 DSM : integer; (* total num blocks -1, max 7FFF *)
                 DRM : integer; (* dir entries -1 *)
                 AL0 : byte;    (* Allocation vector for directory *)
                 AL1 : byte;
                 CKS : integer; (* 8000h=fixed disk *)
                 OFF : integer; (* track offfet=1 *)
                 PSH : byte;    (* physical record shift factor *)
                 PHM : byte;    (* physical record mask *)
                 unused: array[0..488] of byte;
               end;



var Alt_Status,IDE_Data,IDE_Error,
    IDE_SecCnt,IDE_SecNum,IDE_CylLow,
    IDE_CylHigh,IDE_SDH,IDE_CmdStat : integer;

       MBRbuf   : MBRT;
       BootBuf  : BootT;
       i,j,k,l,m       : integer;
       func,c          : char;
       err             : boolean;
       s,s1            : workstr;

(* use our own console status routine, since the one implemented in *)
(* Turbo-Pascal won't detect the "keypressed" status properly.      *)

function ConStat:boolean;
begin
  ConStat:=BIOS(1)>0;
  end;

(* translate numbers into their hex representation (as string). *)

function hexbyte(x:byte):workstr;
const nib : array[0..15] of char = '0123456789ABCDEF';
begin
  hexbyte:=nib[x shr 4]+nib[x and 15];
  end;

function hexword(x:integer):workstr;
begin
  hexword:=hexbyte(hi(x))+hexbyte(lo(x));
  end;

procedure dumpBuffer(var buffer);
var i : integer;
    buf:buftype absolute buffer;
    b:byte;
begin
  s:='';
  write(^L'0000: ');
  for i:=1 to 512 do
  begin
    b:=buf[i-1];
    if (b>=32) and (b<=127) then s:=s+chr(b) else s:=s+'.';
    write(hexbyte(b)+' ');
    if i mod 8 = 0 then write (' ');
    if i mod 16 = 0 then
    begin
      writeln('  ',s); s:='';
      if i < 504 then write (hexword(i)+': ');
      if constat then exit;
    end;
  end;
end;


(* Set the port addresses for the various interface registers. *)
(* The addresses are kept in variables since they can be       *)
(* changed during run-time.                                    *)

procedure SetPorts(var base:integer);
begin
  base:=base and $F0;
  Alt_Status:=base+6;
  IDE_Data:=base+8;
  IDE_Error:=base+9;
  IDE_SecCnt:=base+10;
  IDE_SecNum:=base+11;
  IDE_CylLow:=base+12;
  IDE_CylHigh:=base+13;
  IDE_SDH:=base+14;
  IDE_CmdStat:=base+15;
  writeln('Ports setup for base ',HexByte(base),'h.');
  end;


(* display error status *)

procedure Error(s:workstr; flag:boolean);
begin
  writeln('  ',s,'; Status: ',hexbyte(port[ide_cmdstat]),
          ' ',hexbyte(port[ide_error]));
  if flag then halt;
  end;

(* Wait until the drive is ready to accept a command.       *)
(* The timeout value may be changed according to the drive. *)
(* Remove the "i:=succ(i)" instruction to disable timeout.  *)

procedure wait_ready;
const timeout = 30000;
var i : integer;
begin
  i:=0;
  while (port[ide_cmdstat]>128) and (i<timeout) do i:=succ(i);
  if i=timeout then Error('WaitReady TimeOut',true);
  end;

(* Wait for the drive's Data Request (DRQ). *)
(* For the timeout, see above.              *)

procedure wait_drq;
const timeout = 30000;
var i : integer;
begin
  i:=0;
  while (port[ide_cmdstat] and 8=0) and (i<timeout) do i:=succ(i);
  if i=timeout then Error('WaitDRQ TimeOut',true);
  end;

(* write a command to the drive *)

procedure ide_command(cmd:byte);
begin
  wait_ready;
  port[ide_cmdstat]:=cmd;
  wait_ready;
  end;

(* Read the sector buffer from the drive. *)

function read_secbuf(var buf:buftype):boolean;
var i : integer;
begin
  (* write('ReadBuf press enter'); readln; writeln; *)
  wait_drq;
  for i:=0 to 511 do buf[i]:=port[ide_data];
  read_secbuf:=port[ide_cmdstat] and $89=0;
  end;

(* Write the sector buffer to the drive. *)

function write_secbuf(var buf:buftype):boolean;
var i : integer;
begin
  wait_drq;
  for i:=0 to 511 do port[ide_data]:=buf[i];
  wait_ready;
  write_secbuf:=port[ide_cmdstat] and $89=0;
  end;

(* position the drive on the desired cylinder (seek) *)

function hd_seek(cyl:integer):boolean;
begin
  wait_ready;
  port[ide_cyllow]:=lo(cyl);
  port[ide_cylhigh]:=hi(cyl);
  port[ide_sdh]:=$A0;
  ide_command(cmd_seek);
  hd_seek:=port[ide_cmdstat] and $89=0;
  end;


function hd_readbuf(var buf:buftype):boolean;
begin
  wait_ready;
  ide_command(cmd_readbuffer);
  hd_readbuf:=read_secbuf(buf);
end;

function hd_writebuf(var buf:buftype):boolean;
begin
  wait_ready;
  ide_command(cmd_writebuffer);
  hd_writebuf:=write_secbuf(buf);
end;

(* Read a single sector from the drive. Retry up to 5 times on error. *)
(* Print the number of tries if above 1, and report errors.           *)

procedure hd_readsector(cyl,head,sec:integer; var buf:buftype);
var n : byte;
    b : boolean;
begin
  n:=0;
  repeat
    wait_ready;
    port[ide_error]:=$AA;
    port[ide_seccnt]:=1;
    port[ide_secnum]:=sec;
    port[ide_cyllow]:=lo(cyl);
    port[ide_cylhigh]:=hi(cyl);
    port[ide_sdh]:=$A0+head;
    ide_command(cmd_readsector);
    b:=read_secbuf(buf);
    n:=succ(n);
  until b or (n>5);
  if not b then Error('Read Sector',false);
  end;

(* Write a single sector to the drive. No need for retries yet, *)
(* until now it was just a go/nogo behaviour.                   *)

procedure hd_writesector(cyl,head,sec:integer; var buf:buftype);
begin
  wait_ready;
  port[ide_seccnt]:=1;
  port[ide_secnum]:=sec;
  port[ide_cyllow]:=lo(cyl);
  port[ide_cylhigh]:=hi(cyl);
  port[ide_sdh]:=$A0+head;
  ide_command(cmd_writesector);
  if not write_secbuf(buf) then Error('Write Sector',false);
  end;

(* initialise the harddisk drive and set the desired geometry *)

const hd_InitDone:boolean=false;

procedure hd_init(cyls,hds,secs:integer);
begin
  writeln('Initialising the drive...');
  port[alt_status]:=6;
  delay(10);            (* Drive Software Reset *)
  port[alt_status]:=2;
  wait_ready;
  (* writeln(port[ide_error]:4,port[ide_seccnt]:4,port[ide_secnum]:4,
          port[ide_cyllow]:4,port[ide_cylhigh]:4,port[ide_sdh]:4); *)
  port[ide_seccnt]:=secs;
  port[ide_cyllow]:=lo(cyls);
  port[ide_cylhigh]:=hi(cyls);
  port[ide_sdh]:=pred(hds)+$A0;
  ide_command(cmd_initialize);
  writeln('Mode : ',cyls,'x',hds,'x',secs);
  hd_initDone := true;
  end;



procedure showDPB;
begin
  with BootBuf do
  begin
    WriteLn ('SPT: ',SPT:5,'   128 byte recs per track(all heads)');
    writeln ('BSH: ',BSH:5);
    writeln ('BLM: ',BLM:5);
    writeln ('EXM: ',EXM:5);
    writeln ('DSM: ',DSM:5,'   total num blocks -1');
    writeln ('DRM: ',DRM:5,'   Dir entries');
    writeln ('AL0: ',hexbyte(AL0),'h');
    writeln ('AL1: ',hexbyte(AL1),'h');
    writeln ('CKS: ',hexword(CKS),'h');
    writeln ('OFF: ',OFF:5);
    writeln ('PSH: ',PSH:5);
    writeln ('PHM: ',PHM:5);
  end;
end;


procedure showMBR;
var trkSizeK,sizeK : integer;
    buf : buftype absolute BootBuf;
begin
  with MBRBuf do
  begin
    WriteLn ('Disk geom: Tracks: ',PartData.Tracks,' Sectors: ',PartData.Sectors,' Heads: ',PartData.Heads);
    Writeln ('Boot Partition: ',PartData.BootPartNum);
    TrkSizeK := (PartData.Sectors * PartData.Heads) div 2;
    writeln;
    writeln('Part Start Length Size K   SPT BSH BLM EXM   DSM  DRM AL0 AL1   CKS OFF PSH PHM');
    writeln('-------------------------------------------------------------------------------');
    for i := 0 to 15 do
    if PartData.Partitions[i].Start > 0 then
    begin
      sizeK := PartData.Partitions[i].len * trkSizeK;
      Write (i:4,PartData.Partitions[i].Start:6,
               PartData.Partitions[i].Len:7,sizeK:7);
      if PartData.Partitions[i].Start > 0 then
      begin
        hd_readsector(PartData.Partitions[i].Start,0,1,buf);
        with BootBuf do
        begin
          move(id,s[1],6); s[0]:=#6;
          if s<>'CPM3AD' then
            write('  no CPM partition') else
          begin
            Write (SPT:6,BSH:4,BLM:4,EXM:4,DSM:6,DRM:5,
                   ' ',hexbyte(AL0),'h ',hexbyte(AL1),'h ',
                   hexword (CKS),'h',
                   OFF:4,PSH:4,PHM:4);
          end;
        end;
      end;
      writeln;
    end;
  end;
end;


function readMBR : boolean;
var buf : buftype absolute MBRBuf;
begin
  hd_readsector(0,0,1,buf);
  readMBR := false;
  with MBRBuf do
  begin
    if (id55 <> $55) or (idaa <> $aa) then
    begin
      writeln ('no valid end Signature found'); exit;
    end;
    if (PartData.CpmId[0] <> 'C') or
       (PartData.CpmId[1] <> 'P') or
       (PartData.CpmId[2] <> 'M') then
    begin
      writeln ('MBR does not contain a vaild CPM signature'); exit;
    end;
    cylinders   := PartData.Tracks;
    heads       := PartData.Heads;
    sectors     := PartData.Sectors;
    if not hd_InitDone then
      hd_init(cylinders,heads,sectors);
  end;
  readMBR := true;
end;

procedure writeMBR;
var buf : buftype absolute MBRBuf;
begin
  hd_writesector(0,0,1,buf);
  writeln ('MBR written to disk');
end;


procedure DeletePartition;
var Part : byte;
    Bbuf : buftype absolute BootBuf;
    Mbuf : buftype absolute MBRBuf;
begin
  if not readMBR then exit;
  ShowMBR;
  Write('Partition to delete: '); readln(Part);
  if MBRBuf.PartData.Partitions[part].Start > 0 then
  begin
    write('To delete Partition ',Part,' enter YES: '); readln(s);
    if s = 'YES' then
    begin
      FillChar(BBuf,sizeof(BBuf),$e5);
      hd_writesector(MBRBuf.PartData.Partitions[part].Start,0,1,BBuf);
      fillchar(MBRBuf.PartData.Partitions[Part],sizeof(MBRBuf.PartData.Partitions[Part]),0);
      hd_writesector(0,0,1,MBuf);
      writeln('Partition Deleted');
    end else writeln ('Delete aborted');
  end else writeln ('Invalid or unused Partition specified');
end;


procedure addPartition;
var FreePart : byte;
    FreeTrk,BlocksDir,BlocksTrk,SectorsTrk,Trks,BlockSize,DirE: integer;
    buf : buftype absolute BootBuf;
begin
  if not readMBR then exit;
  FreePart := $FF;
  i := 0;
  while (i<16) and (FreePart = $FF) do
  begin
    if MBRBuf.PartData.Partitions[i].start = 0 then
      FreePart := i;
    i:=succ(i);
  end;
  if FreePart = $FF then
  begin
    writeln ('All 16 partitions are already allocated'); exit;
  end;
  FillChar(BootBuf,sizeof(BootBuf),0);
  with BootBuf do
  begin
    s:='CPM3AD';
    Move(s[1],id,6);
    Trks := 0;
    write ('Number of tracks to allocate: '); readln(Trks);
    if Trks < 1 then exit;
    repeat
      write ('Blocksize in bytes: '); readln (Blocksize);
      if Blocksize = 0 then exit;
    until (BlockSize=1024) or (BlockSize=2048) or (BlockSize=4096) or
          (BlockSize=8192) or (BlockSize=16384);
    repeat
      write ('Number of directory entries: '); readln (DirE);
      if DirE < 1 then exit;
    until DirE div (BlockSize div 32) <= 16;
    SectorsTrk := Sectors * heads;
    BlocksTrk := SectorsTrk div (BlockSize div 512);
    SPT := SectorsTrk * 4;  (* 128 byte recs per track(all heads) *)
    Case BlockSize OF
      1024 : begin BSH := 3; BLM := 7; end;  (* system guide page 42 *)
      2048 : begin BSH := 4; BLM := 15; end;
      4096 : begin BSH := 5; BLM := 31; end;
      8192 : begin BSH := 6; BLM := 63; end;
      16384: begin BSH := 7; BLM := 127; end;
    end;
    DSM := (BlocksTrk * Trks) -1; (* total num blocks -1, max 7FFF *)
    Case BlockSize of
      1024 : if DSM < 256 then EXM := 0 else begin writeln ('disksize not allowed with 1K blocks'); exit; end;
      2048 : if DSM < 256 THEN EXM := 1 else EXM := 0;
      4096 : if DSM < 256 then EXM := 3 else EXM := 1;
      8192 : if DSM < 256 then EXM := 7 else EXM := 3;
      16384: if DSM < 256 then EXM := 15 else EXM := 7;
    end;
    DRM := DirE-1;  (* dir entries -1 *)

    BlocksDir := DirE div (BlockSize div 32);
    if BlocksDir * BlockSize div 32 < DirE then BlocksDir := BlocksDir +1;
    writeln ('BlocksDir: ',BlocksDir);
    AL0 := 0; AL1 := 0;
    j := $8000;
    l := BlocksDir;
    while (l>0) do
    begin
      AL0 := AL0 or hi(j);
      AL1 := AL1 or lo(j);
      l := pred(l);
      j := j shr 1;
    end;
    CKS := $8000;  (* 8000h=fixed disk *)
    OFF := 1;      (* track offfet=1 *)
    PSH := 2;      (* physical record shift factor *)
    PHM := 3;      (* physical record mask *)
    showDPB;
    write ('Ok to add this Partition with ',Trks,' Tracks ?'); readln(s);
    if upcase(copy(s,1,1)) = 'Y' then
    begin
      FreeTrk := 1;
      for i := 0 to 15 do
        if MBRBuf.PartData.Partitions[i].Start <> 0 then
        begin
          j := MBRBuf.PartData.Partitions[i].Start +MBRBuf.PartData.Partitions[i].Len;
          if FreeTrk < j then FreeTrk := j;
        end;
      writeln ('Free from track ',FreeTrk);
      if FreeTrk+Trks>Cylinders then
      begin
        writeln ('not enough free space after last partition');
        exit;
      end;
      MBRBuf.PartData.Partitions[FreePart].Start := FreeTrk;
      MBRBuf.PartData.Partitions[FreePart].Len := Trks;
      hd_writesector(FreeTrk,0,1,buf);
      Writeln ('1st sector of partition written');
      WriteMBR;
      (* initialize the directory *)
      j := FreeTrk+1;
      k := 0;
      l := 1;
      fillchar(buf,sizeof(buf),$E5);
      Write ('Init Directory ');
      for i:=1 to (DirE div 16)+1 do (* one more *)
      begin
        (* writeln ('Init Dir Trk:',j,' Head:',k,' Sector:',l); *)
        hd_writesector(j,k,l,buf);
        Write('.');
        l := succ(l);
        if l > sectors then
        begin
          l := 1;
          k := succ(k);
          if k>heads-1 then
          begin
            j:=succ(j);
            k:=0;
          end;
        end;
      end;
      writeln;
    end;
  end;
end;

procedure SetBootPartition;
var Part : byte;
begin
  if not readMBR then exit;
  ShowMBR;
  Write('Boot Partition: '); readln(Part);
  if MBRBuf.PartData.Partitions[part].Start > 0 then
  begin
    MBRBuf.PartData.BootPartNum := Part;
    Writeln ('Part ',Part,' active');
    writeMBR;
  end else writeln ('Invalid or unused Partition specified');
end;

(*$I-*)
procedure UpdateMBRCode;
var f : file;
    Bbuf : buftype absolute BootBuf;
begin
  if not readMBR then exit;
  s:='';
  write ('File name for MBR boot code (IDEBOOT.BIN): '); readln(s);
  if s = '' then s := 'IDEBOOT.BIN';
  write ('Reading ',s,' ');
  assign(F,s); reset(f);
  if IOResult = 0 then
  begin
    j:=0;
    FillChar(MBRBuf.BootCode,sizeof(MBRBuf.BootCode),0);

    for i := 1 to 3 do
    begin
      BlockRead(f,Bbuf[j],1);
      j:=j+128;
      write('.');
    end;
    writeln;
    Move(BBuf,MBRBuf.BootCode,sizeof(MBRBuf.BootCode));
    writeMBR;
  end else
    writeln ('unable to open ',s);
end;

procedure UpdatePartitionBootCode;
var f : file;
    Bbuf : buftype absolute BootBuf;
    Trk,Count:integer;
    Part:byte;
    fend:boolean;
begin
  if not readMBR then exit;
  ShowMBR;
  Write('Partition: '); readln(Part);
  trk := MBRBuf.PartData.Partitions[part].Start;
  if trk > 0 then
  begin
    write ('File name for Partition boot code (IDELDR.BIN): '); readln(s);
    if s = '' then s := 'IDELDR.BIN';
    assign(f,s); reset(f); count := 0; fend:=false;
    if IOResult = 0 then
    begin
      j:=0; k:=2;
      while true do
      begin
        fillchar(BBuf,sizeof(Bbuf),$e5);
        for i := 0 to 3 do
          if not fend then
          begin
            BlockRead(f,BBuf[i*128],1);
            fend := (IOResult <> 0);
          end;
        hd_writesector(Trk,j,k,Bbuf);
        count:=succ(count);
        write('.');
        if fend then
        begin
          writeln(' EOF, ',count,' sectors written');
          close(f); exit;
        end;
        k:=succ(k);
        if k > Sectors then
        begin
          k:=1; j:=succ(j);
          if j>heads then
          begin
            writeln (' Boottrack full, ',count,' sectors written');
            close(f); exit;
          end;
        end;
      end;
    end else writeln ('Unable to open ',s);
  end else writeln ('Invalid or unused Partition specified');
end;


procedure InitMBR;
begin
  write ('WARNING: Initializing the MBR will result in all data lost, to continue enter YES: ');
  readln(s);
  if s<>'YES' then exit;
  fillchar(MBRBuf,sizeof(MBRBuf),0);
  Write ('Enter Disk number of Tracks (',cylinders,'): '); readln(cylinders);
  write ('Enter Disk number of heads (',heads,'): '); readln (heads);
  write ('Enter Disk number of sectors per track (',sectors,'): '); readln(sectors);
  writeln;
  hd_init(cylinders,heads,sectors);
  writeln;
  write ('Continue with above parameters ? (YES or ENTER for no)'); readln(s);
  if s <> 'YES' then exit;
  with MBRBuf do
  begin
    Id55 := $55; IdAA := $AA;
    PartData.cpmId[0] := 'C';
    PartData.cpmId[1] := 'P';
    PartData.cpmId[2] := 'M';
    PartData.Tracks := cylinders;
    PartData.Heads := heads;
    PartData.Sectors := sectors;
  end;
  writeMBR;
end;

procedure DumpMBR;
begin
  if not readMBR then exit;
  dumpBuffer(MBRBuf);
  readln;
end;

procedure DumpPartition;
var Part : byte;
    Bbuf : buftype absolute BootBuf;
    Trk:integer;
    Sec,Head:byte;
    c:char;
begin
  if not readMBR then exit;
  ShowMBR;
  Write('Partition: '); readln(Part);
  Trk := MBRBuf.PartData.Partitions[part].Start;
  sec:=1; Head:=0;
  if trk > 0 then
  repeat
    hd_readsector(Trk,Head,Sec,Bbuf);
    dumpBuffer(Bbuf);
    write('T: ',Trk,' H: ',Head,' S: ',Sec,'  t:next,T:prev Trk, space or n: next, N:prev sec, x:exit ');
    read(kbd,c);
    case c of
      ' ',
      'n' : begin
              Sec:=succ(Sec); if sec>Sectors then
              begin
                Sec:=1;
                if Sec>Sectors then
                begin
                  Sec:=1; Trk := succ(Trk);
                end;
              end;
            end;
      'N' : begin
              Sec:=pred(Sec); if sec=0 then
              begin
                sec:=Sectors;
                if head>0 then head:=pred(head) else
                begin
                  head := pred(Heads);
                  if trk>0 then trk:=pred(Trk);
                end;
              end;
            end;
      't' : if trk < Cylinders then trk:=succ(trk);
      'T' : if trk > 0 then trk := pred(Trk);
    end;
  until (c='x') or (c='X');
  writeln;
end;


(* MAIN *)

begin
  constptr:=addr(constat);
  writeln(signon);
  SetPorts(GIDEbase);
  writeln(sizeof(MBRT),' ',sizeof(BootT));
  repeat
    write(^m^j'Functions:'^m^j,
    '(0) Initialise MBR (All Data will be lost)'^m^j,
    '(1) Show partition data'^m^j,
    '(2) Add Partition'^m^j,
    '(3) Delete Partition'^m^j,
    '(4) Set Boot Partition'^m^j,
    '(5) Update MBR code from IDEBOOT.BIN'^m^j,
    '(6) Update Partition boot code from IDELDR.BIN'^m^j,
    '(7) Dump MBR'^m^j,
    '(8) Dump Partition'^m^j,
    '(X) Exit'^m^j,
    'Input: ');
    repeat read(kbd,func); func:=upcase(func)
    until func in ['0'..'8','X'];
    write(func,^m^j^m^j);
    case func of
      '0' : InitMBR;
      '1' : if ReadMBR then ShowMBR;
      '2' : addPartition;
      '3' : deletePartition;
      '4' : setBootPartition;
      '5' : UpdateMBRCode;
      '6' : UpdatePartitionBootCode;
      '7' : DumpMBR;
      '8' : DumpPartition;
    end;
  until func='X';
  end.
