	title	'Root module of Loader BIOS for CP/M 3.0'

; Unnecessary BIOS functions can be deleted from the LDRBIOS to conserve space. There
; is one absolute restriction on the length of the LDRBIOS; it cannot extend above the base of the
; banked portion of CP/M 3. (GENCPM lists the base address of CP/M 3 in its load map.) If you
; plan to boot CP/M 3 from standard, single-density, eight-inch floppy disks, your CPMLDR must
; not be longer than 1980H to place the CPMLDR.COM file on two system tracks with the boot
; sector. If the CCP resides on the system tracks with the Cold Start Loader and CPMLDR, the
; combined lengths must not exceed 1980H.


true	equ -1
false	equ not true

banked	equ false

	maclib z80

cr	equ 13
lf	equ 10
bell	equ 7
ctlQ	equ 'Q'-'@'
ctlS	equ 'S'-'@'


    ; variables in system data page

	extrn intser
	extrn monoff
	
    ; initialization

	;	extrn ?init			; general initialization and signon
	;	extrn binit			; baudraten init fuer PROF-80

    ; user defined character I/O routines

	extrn ?ci,?co,?cist,?cost	; each take device in <B>
	extrn ?cinit			; (re)initialize device in <C>
	extrn @ctbl			; physical character device table

    ; disk communication data items

	public @adrv,@rdrv,@trk,@sect	; parameters for disk I/O
	public @dma,@dbnk,@cnt		;    ''       ''   ''  ''

    ; memory control

	public @cbnk			; current bank

    ; clock support


    ; general utility routines

	public ?pmsg,?pdec	; print message, print number from 0 to 65535
	public ?pderr		; print BIOS disk error message header

	maclib modebaud		; define mode bits
	maclib syspage
	maclib cdef


    ; External names for BIOS entry points

	public ?boot,?wboot,?const,?conin,?cono,?list,?auxo,?auxi
	public ?home,?sldsk,?sttrk,?stsec,?stdma,?read,?write
	public ?lists,?sctrn
	public ?conos,?auxis,?auxos,?dvtbl,?devin,?drtbl
	public ?mltio,?flush,?mov,?tim,?bnksl,?stbnk,?xmov


    ; BIOS Jump vector.

		; All BIOS routines are invoked by calling these
		;	entry points.

?boot:	jmp boot	; initial entry on cold start
?wboot:	jmp wboot	; reentry on program exit, warm start

?const:	jmp const	; return console input status
?conin:	jmp conin	; return console input character
?cono:	jmp conout	; send console output character
?list:	jmp dummy	; send list output character
?auxo:	jmp dummy	; send auxilliary output character
?auxi:	jmp dummy	; return auxilliary input character

?home:	jmp home	; set disks to logical home
?sldsk:	jmp seldsk	; select disk drive, return disk parameter info
?sttrk:	jmp settrk	; set disk track
?stsec:	jmp setsec	; set disk sector
?stdma:	jmp setdma	; set disk I/O memory address
?read:	jmp read	; read physical block(s)
?write:	jmp dummy	; write physical block(s)

?lists:	jmp dummy	; return list device status
?sctrn:	jmp sectrn	; translate logical to physical sector

?conos:	jmp conost	; return console output status
?auxis:	jmp dummy	; return aux input status
?auxos:	jmp dummy	; return aux output status
?dvtbl:	jmp devtbl	; return address of device def table
?devin:	jmp dummy	; change baud rate of device

?drtbl:	jmp getdrv	; return address of disk drive table
?mltio:	jmp multio	; set multiple record count for disk I/O
?flush:	jmp flush	; flush BIOS maintained disk caching

?mov:	jmp move	; block move memory to memory
?tim:	jmp dummy	; ?time	; Signal Time and Date operation
?bnksl:	jmp dummy	; select bank for code execution and default DMA
?stbnk:	jmp dummy	; select different bank for disk I/O DMA operations.
?xmov:	jmp dummy	; set source and destination banks for one operation

	jmp 0		; reserved for future expansion
	jmp 0		; reserved for future expansion
	jmp 0		; reserved for future expansion


			; table of pointers to XDPHs
	public @dtbl
	extrn fdsd0

@dtbl	dw fdsd0
	dw 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0	; drives B-P non-existant

	title 'System Control Block Definition for CP/M3 BIOS'

	public @civec, @covec, @aivec, @aovec, @lovec
;, @bnkbf
	public @crdma, @crdsk, @vinfo, @resel, @fx, @usrcd 
        public @mltio


@CIVEC  dw 0   ; Console Input Redirection 
                                ; Vector (word, r/w)
@COVEC  dw 0    ; Console Output Redirection 
                                ; Vector (word, r/w)
@AIVEC  dw 0    ; Auxiliary Input Redirection 
                                ; Vector (word, r/w)
@AOVEC  dw 0    ; Auxiliary Output Redirection 
                                ; Vector (word, r/w)
@LOVEC  dw 0    ; List Output Redirection 
                                ; Vector (word, r/w)
;@BNKBF  dw 0    ; Address of 128 Byte Buffer 
                                ; for Banked BIOS (word, r/o)
@CRDMA  dw 0   ; Current DMA Address 
                                ; (word, r/o)
@CRDSK  dw 0   ; Current Disk (byte, r/o)
@VINFO  dw 0   ; BDOS Variable "INFO" 
                                ; (word, r/o)
@RESEL  dw 0   ; FCB Flag (byte, r/o)
@FX     dw 0   ; BDOS Function for Error 
                                ; Messages (byte, r/o)
@USRCD  dw 0   ; Current User Code (byte, r/o)
@MLTIO	dw 0	; Current Multi-Sector Count
				; (byte,r/w)

	; BOOT
	;	Initial entry point for system startup.

boot:
	;	lxi sp,boot$stack
;	call EPROM$OFF2	; Monitor aus
	call binit	; setzte  duplex baud fuer PROF-80
	mvi c,15	; initialize all 16 character devices
c$init$loop:
	push b ! call ?cinit ! pop b
	dcr c ! jp c$init$loop

	call ?init	; perform any additional system initialization
			; and print signon message

boot$1:
	ret


	; WBOOT
	;	Entry for system restarts.

wboot:
	call set$jumps		; initialize page zero, AD: needed ??
	ret

set$jumps:

  if banked
	mvi a,1 ! call ?bnksl
  endif
	ret



EPROM$OFF2:
  	MVI	C,P$RAM$TABLE	; EPROM AUSBLENDEN
	MVI	E,02H
	MVI	B,00H
	OUTP	E
	MVI	B,10H
	OUTP	E
	MVI	A,0
	STA	EPROMZ
	RET



;	mvi a,JMP
;	sta 0 ! sta 5		; set up jumps in page zero
;	lxi h,?wboot ! shld 1	; BIOS warm start entry
;	lhld @MXTPA ! shld 6	; BDOS system call entry
	ret

binit:
	lda	baud$jumperbyte		; setze duplex baudrate
	mov	l,a ! mvi h,0		; entsprechend j4
	lxi	d,j4baudlist ! dad d	;
	mov	a,m ! sta @ctbl+7	; das sollte besser ein symbol sein, passt aber
	ret	


j4baudlist:
	db	baud$9600
	db	baud$4800
	db	baud$2400
	db	baud$1200
	db	baud$300

	
	; diese Routinen werden vom EPROM benutzt

xwboot:
	mvi a,1 ! sta epromz
	call monoff ! jmp ?wboot

xconst:
	push b ! push d ! push h
	call ?const
	pop h ! pop d ! pop b ! ret

xconin:
	push b ! push d ! push h
	call ?conin
	pop h ! pop d ! pop b ! ret

xcono:
	push b ! push d ! push h
	call ?cono
	pop h ! pop d ! pop b ! ret



	; DEVTBL
	;	Return address of character device table

devtbl:
	lxi h,@ctbl ! ret


	; GETDRV
	;	Return address of drive table

getdrv:
	lxi h,@dtbl ! ret



	; CONOUT
	;	Console Output.  Send character in <C>
	;			to all selected devices

conout:	

	lhld @covec	; fetch console output bit vector
	jmp out$scan



out$scan:
	mvi b,0		; start with device 0
co$next:
	dad h		; shift out next bit
	jnc not$out$device
	push h		; save the vector
	push b		; save the count and character
not$out$ready:
	call coster ! ora a ! jz not$out$ready
	pop b ! push b	; restore and resave the character and device
	call ?co	; if device selected, print it
	pop b		; recover count and character
	pop h		; recover the rest of the vector
not$out$device:
	inr b		; next device number
	mov a,h ! ora l	; see if any devices left
	jnz co$next	; and go find them...
	ret


	; CONOST
	;	Console Output Status.  Return true if
	;		all selected console output devices
	;		are ready.

conost:
	lhld @covec	; get console output bit vector
	jmp ost$scan



ost$scan:
	mvi b,0		; start with device 0
cos$next:
	dad h		; check next bit
	push h		; save the vector
	push b		; save the count
	mvi a,0FFh	; assume device ready
	cc coster	; check status for this device
	pop b		; recover count
	pop h		; recover bit vector
	ora a		; see if device ready
	rz		; if any not ready, return false
	inr b		; drop device number
	mov a,h ! ora l	; see if any more selected devices
	jnz cos$next
	ori 0FFh	; all selected were ready, return true
	ret

coster:		; check for output device ready, including optional
		;	xon/xoff support
	mov l,b ! mvi h,0	; make device code 16 bits
	push h			; save it in stack
	dad h ! dad h ! dad h	; create offset into device characteristics tbl
	lxi d,@ctbl+6 ! dad d	; make address of mode byte
	mov a,m ! ani mb$xonxoff
	pop h			; recover console number in <HL>
	jz ?cost		; not a xon device, go get output status direct
	lxi d,xofflist ! dad d	; make pointer to proper xon/xoff flag
	call cist1		; see if this keyboard has character
	mov a,m ! cnz ci1	; get flag or read key if any
	cpi ctlq ! jnz not$q	; if its a ctl-Q,
	mvi a,0FFh 		;	set the flag ready
not$q:
	cpi ctls ! jnz not$s	; if its a ctl-S,
	mvi a,00h		;	clear the flag
not$s:
	mov m,a			; save the flag
	call cost1		; get the actual output status,
	ana m			; and mask with ctl-Q/ctl-S flag
	ret			; return this as the status

cist1:			; get input status with <BC> and <HL> saved
	push b ! push h 
	call ?cist
	pop h ! pop b
	ora a
	ret

cost1:			; get output status, saving <BC> & <HL>
	push b ! push h
	call ?cost
	pop h ! pop b
	ora a
	ret

ci1:			; get input, saving <BC> & <HL>
	push b ! push h
	call ?ci
	pop h ! pop b
	ret


	; CONST
	;	Console Input Status.  Return true if
	;		any selected console input device
	;		has an available character.

const:
	lhld @civec	; get console input bit vector
	;	jmp ist$scan

ist$scan:
	mvi b,0		; start with device 0
cis$next:
	dad h		; check next bit
	mvi a,0		; assume device not ready
	cc cist1	; check status for this device
	ora a ! rnz	; if any ready, return true
	inr b		; drop device number
	mov a,h ! ora l	; see if any more selected devices
	jnz cis$next
	xra a		; all selected were not ready, return false
	ret


	; CONIN
	;	Console Input.  Return character from first
	;		ready console input device.

conin:
	lhld @civec
	jmp in$scan


	; AUXIN
	;	Auxiliary Input.  Return character from first
	;		ready auxiliary input device.


in$scan:
	push h		; save bit vector
	mvi b,0
ci$next:
	dad h		; shift out next bit
	mvi a,0		; insure zero a  (nonexistant device not ready).
	cc cist1	; see if the device has a character
	ora a
	jnz ci$rdy	; this device has a character
	inr b		; else, next device
	mov a,h ! ora l	; see if any more devices
	jnz ci$next	; go look at them
	pop h		; recover bit vector
	jmp in$scan	; loop til we find a character

ci$rdy:
	pop h		; discard extra stack
	jmp ?ci


;	Utility Subroutines

ipchl:		; vectored CALL point
	pchl

?pmsg:		; print message @<HL> up to a null
		; saves <BC> & <DE>
	push b
	push d
pmsg$loop:
	mov a,m ! ora a ! jz pmsg$exit
	mov c,a ! push h
	call ?cono ! pop h
	inx h ! jmp pmsg$loop
pmsg$exit:
	pop d
	pop b
	ret

?pdec:		; print binary number 0-65535 from <HL>
	lxi b,table10! lxi d,-10000
next:
	mvi a,'0'-1
pdecl:
	push h! inr a! dad d! jnc stoploop
	inx sp! inx sp! jmp pdecl
stoploop:
	push d! push b
	mov c,a! call ?cono
	pop b! pop d
nextdigit:
	pop h
	ldax b! mov e,a! inx b
	ldax b! mov d,a! inx b
	mov a,e! ora d! jnz next
	ret

table10:
	dw	-1000,-100,-10,-1,0

?pderr:
	lxi h,drive$msg ! call ?pmsg			; error header
	lda @adrv ! adi 'A' ! mov c,a ! call ?cono	; drive code
	lxi h,track$msg ! call ?pmsg			; track header
	lhld @trk ! call ?pdec				; track number
	lxi h,sector$msg ! call ?pmsg			; sector header
	lhld @sect ! call ?pdec				; sector number
	ret


xofflist	db	-1,-1,-1,-1,-1,-1,-1,-1		; ctl-s clears to zero
		db	-1,-1,-1,-1,-1,-1,-1,-1




;	Disk I/O interface routines


	; SELDSK
	;	Select Disk Drive.  Drive code in <C>.
	;		Invoke login procedure for drive
	;		if this is first select.  Return
	;		address of disk parameter header
	;		in <HL>

seldsk:
	mov a,c ! sta @adrv			; save drive select code
	mov l,c ! mvi h,0 ! dad h		; create index from drive code
	lxi b,@dtbl ! dad b			; get pointer to dispatch table
	mov a,m ! inx h ! mov h,m ! mov l,a	; point at disk descriptor
	ora h ! rz 				; if no entry in table, no disk
	mov a,e ! ani 1 ! jnz not$first$select	; examine login bit
	push h ! xchg				; put pointer in stack & <DE>
	lxi h,-2 ! dad d ! mov a,m ! sta @RDRV	; get relative drive
	lxi h,-6 ! dad d			; find LOGIN addr
	mov a,m ! inx h ! mov h,m ! mov l,a	; get address of LOGIN routine
	call ipchl				; call LOGIN
	pop h					; recover DPH pointer
not$first$select:
	ret


	; HOME
	;	Home selected drive.  Treated as SETTRK(0).

home:
	lxi b,0		; same as set track zero


	; SETTRK
	;	Set Track. Saves track address from <BC> 
	;		in @TRK for further operations.

settrk:
	mov l,c ! mov h,b
	shld @trk
	ret


	; SETSEC
	;	Set Sector.  Saves sector number from <BC>
	;		in @sect for further operations.

setsec:
	mov l,c ! mov h,b
	shld @sect
	ret


	; SETDMA
	;	Set Disk Memory Address.  Saves DMA address
	;		from <BC> in @DMA and sets @DBNK to @CBNK
	;		so that further disk operations take place
	;		in current bank.

setdma:
	mov l,c ! mov h,b
	shld @dma

	lda @cbnk	; default DMA bank is current bank
			; fall through to set DMA bank

dummy:
	ret


	; SECTRN
	;	Sector Translate.  Indexes skew table in <DE>
	;		with sector in <BC>.  Returns physical sector
	;		in <HL>.  If no skew table (<DE>=0) then
	;		returns physical=logical.

sectrn:
	mov l,c ! mov h,b
	mov a,d ! ora e ! rz
	xchg ! dad b ! mov l,m ! mvi h,0
	ret


	; READ
	;	Read physical record from currently selected drive.
	;		Finds address of proper read routine from
	;		extended disk parameter header (XDPH).

read:
	lhld @adrv ! mvi h,0 ! dad h	; get drive code and double it
	lxi d,@dtbl ! dad d		; make address of table entry
	mov a,m ! inx h ! mov h,m ! mov l,a	; fetch table entry
	push h				; save address of table
	lxi d,-8 ! dad d		; point to read routine address
	;	jmp rw$common			; use common code


rw$common:
	mov a,m ! inx h ! mov h,m ! mov l,a	; get address of routine
	pop d				; recover address of table
	dcx d ! dcx d			; point to relative drive
	ldax d ! sta @rdrv		; get relative drive code and post it
	inx d ! inx d			; point to DPH again
	pchl				; leap to driver


	; MULTIO
	;	Set multiple sector count. Saves passed count in
	;		@CNT


	; MULTIO
	;	Set multiple sector count. Saves passed count in
	;		@CNT

multio:	ret


	; FLUSH
	;	BIOS deblocking buffer flush.  Not implemented.

flush:	xra a ! ret		; return with no error

move:
	xchg			; we are passed source in DE and dest in HL
	ldir			; use Z80 block move instruction
	xchg			; need next addresses in same regs
	ret


;--------------- Init Stuff from boot.asm ------------------------

?init:
	xra	a ! sta epromz				;
	mvi	a,com$page ! sta common$page		;
	mvi	a,0c3h ! sta 38h			; interuppt in
	lxi	h,intser ! shld 39h			; bank 0

	lxi h,xwboot ! shld umleit$getcm + 1		; setze umleit
	lxi h,xconst ! shld umleit$const + 1		; vektoren
	lxi h,xconin ! shld umleit$conin + 1		;
	lxi h,xcono ! shld umleit$conout + 1		;

	lda	io$jumperbyte		; waehle richtige Voreinstellung
	cpi	4 		; der Geraetetreiber
	jz iniio4
	cpi	3		; in Abhaengigkeit des IO Jumpers
	jz iniio3
	cpi	2
	jz iniio2
	cpi	1
	jz iniio1

iniio0:	lxi h,00800h ! lxi d,00400h ! lxi b,08000h	; j4 kein jumper 
	jmp ioset
iniio1:	lxi h,08000h ! lxi d,04000h ! lxi b,08000h	; j4/1-3
	jmp ioset
iniio2:	lxi h,02000h ! lxi d,04000h ! lxi b,08000h	; j4/2-4
	jmp ioset
iniio3:	lxi h,01000h ! lxi d,04000h ! lxi b,08000h	; j4/3-5
	jmp ioset
iniio4:	lxi h,00800h ! lxi d,00400h ! lxi b,08000h	; j4/4-6

ioset:	shld @civec
	shld @covec
;	sded @lovec
;	sbcd @aivec
;	sbcd @aovec
	lxi h,signon$msg ! call ?pmsg			; print signon message
	ret

signon$msg:	db	13,10
		db	'PROF-80 CP/M3 LDR Bios AD '
		maclib  makedate
		@BDATE					; Build date

		db	13,10,0

;-----------------------------------------------------------------


	; error message components
drive$msg	db	cr,lf,bell,'LDRBIOS Err on ',0
track$msg	db	': T-',0
sector$msg	db	', S-',0


    ; disk communication data items

@adrv	ds	1		; currently selected disk drive
@rdrv	ds	1		; controller relative disk drive
@trk	ds	2		; current track number
@sect	ds	2		; current sector number
@dma	ds	2		; current DMA address
@cnt	db	0		; record count for multisector transfer
@dbnk	db	0		; bank for DMA operations
@cbnk	db	0		; bank for processor operations


	end
